@@ -89,17 +89,20 @@ Member Documentation
 ===============================
 """
 from __future__ import absolute_import
+from __future__ import division
 
 import os
 import random
 import re
 import shlex
+import six
 import tempfile
 import time
 
 from pwnlib import adb
 from pwnlib import atexit
 from pwnlib import elf
+from pwnlib import qemu
 from pwnlib import tubes
 from pwnlib.asm import _bfdname
 from pwnlib.asm import make_elf
@@ -107,7 +110,6 @@ from pwnlib.asm import make_elf_from_assembly
 from pwnlib.context import LocalContext
 from pwnlib.context import context
 from pwnlib.log import getLogger
-from pwnlib.qemu import get_qemu_user
 from pwnlib.util import misc
 from pwnlib.util import proc
 
@@ -115,7 +117,7 @@ log = getLogger(__name__)
 
 @LocalContext
 def debug_assembly(asm, gdbscript=None, vma=None):
-    """debug_assembly(asm, gdbscript=None, vma=None) -> tube
+    r"""debug_assembly(asm, gdbscript=None, vma=None) -> tube
 
     Creates an ELF file, and launches it under a debugger.
 
@@ -127,13 +129,20 @@ def debug_assembly(asm, gdbscript=None, vma=None):
         asm(str): Assembly code to debug
         gdbscript(str): Script to run in GDB
         vma(int): Base address to load the shellcode at
-        **kwargs: Override any :obj:`.context` values.
+        \**kwargs: Override any :obj:`pwnlib.context.context` values.
 
     Returns:
         :class:`.process`
+
+    Example:
+
+    >>> assembly = shellcraft.echo("Hello world!\n")
+    >>> io = gdb.debug_assembly(assembly)
+    >>> io.recvline()
+    b'Hello world!\n'
     """
     tmp_elf = make_elf_from_assembly(asm, vma=vma, extract=False)
-    os.chmod(tmp_elf, 0777)
+    os.chmod(tmp_elf, 0o777)
 
     atexit.register(lambda: os.unlink(tmp_elf))
 
@@ -146,22 +155,30 @@ def debug_assembly(asm, gdbscript=None, vma=None):
 
 @LocalContext
 def debug_shellcode(data, gdbscript=None, vma=None):
-    """
+    r"""debug_shellcode(data, gdbscript=None, vma=None) -> tube
     Creates an ELF file, and launches it under a debugger.
 
     Arguments:
         data(str): Assembled shellcode bytes
         gdbscript(str): Script to run in GDB
         vma(int): Base address to load the shellcode at
-        **kwargs: Override any :obj:`.context` values.
+        \**kwargs: Override any :obj:`pwnlib.context.context` values.
 
     Returns:
         :class:`.process`
+
+    Example:
+
+    >>> assembly = shellcraft.echo("Hello world!\n")
+    >>> shellcode = asm(assembly)
+    >>> io = gdb.debug_shellcode(shellcode)
+    >>> io.recvline()
+    b'Hello world!\n'
     """
-    if isinstance(data, unicode):
+    if isinstance(data, six.text_type):
         log.error("Shellcode is cannot be unicode.  Did you mean debug_assembly?")
     tmp_elf = make_elf(data, extract=False, vma=vma)
-    os.chmod(tmp_elf, 0777)
+    os.chmod(tmp_elf, 0o777)
 
     atexit.register(lambda: os.unlink(tmp_elf))
 
@@ -172,8 +189,8 @@ def debug_shellcode(data, gdbscript=None, vma=None):
 
     return debug(tmp_elf, gdbscript=gdbscript, arch=context.arch)
 
-def _gdbserver_args(pid=None, path=None, args=None, which=None):
-    """_gdbserver_args(pid=None, path=None) -> list
+def _gdbserver_args(pid=None, path=None, args=None, which=None, env=None):
+    """_gdbserver_args(pid=None, path=None, args=None, which=None, env=None) -> list
 
     Sets up a listening gdbserver, to either connect to the specified
     PID, or launch the specified binary by its full path.
@@ -210,7 +227,7 @@ def _gdbserver_args(pid=None, path=None, args=None, which=None):
 
     orig_args = args
 
-    gdbserver_args = [gdbserver]
+    gdbserver_args = [gdbserver, '--multi']
     if context.aslr:
         gdbserver_args += ['--no-disable-randomization']
     else:
@@ -219,6 +236,14 @@ def _gdbserver_args(pid=None, path=None, args=None, which=None):
     if pid:
         gdbserver_args += ['--once', '--attach']
 
+    if env:
+        env_args = []
+        for key in tuple(env):
+            if key.startswith('LD_'): # LD_PRELOAD / LD_LIBRARY_PATH etc.
+                env_args.append('{}={}'.format(key, env.pop(key)))
+        if env_args:
+            gdbserver_args += ['--wrapper', 'env'] + env_args + ['--']
+
     gdbserver_args += ['localhost:0']
     gdbserver_args += args
 
@@ -230,10 +255,16 @@ def _gdbserver_port(gdbserver, ssh):
     # Process /bin/bash created; pid = 14366
     # Listening on port 34816
     process_created = gdbserver.recvline()
+
+    if process_created.startswith(b'ERROR:'):
+        raise ValueError(
+            'Failed to spawn process under gdbserver. gdbserver error message: %s' % process_created
+        )
+
     gdbserver.pid   = int(process_created.split()[-1], 0)
 
-    listening_on = ''
-    while 'Listening' not in listening_on:
+    listening_on = b''
+    while b'Listening' not in listening_on:
         listening_on    = gdbserver.recvline()
 
     port = int(listening_on.split()[-1])
@@ -249,7 +280,7 @@ def _gdbserver_port(gdbserver, ssh):
         listener.level = 'error'
 
         # Hook them up
-        remote <> listener
+        remote.connect_both(listener)
 
     # Set up port forwarding for ADB
     elif context.os == 'android':
@@ -268,8 +299,8 @@ def _get_runner(ssh=None):
     else:                          return tubes.process.process
 
 @LocalContext
-def debug(args, gdbscript=None, exe=None, ssh=None, env=None, **kwargs):
-    """debug(args) -> tube
+def debug(args, gdbscript=None, exe=None, ssh=None, env=None, sysroot=None, **kwargs):
+    r"""debug(args) -> tube
 
     Launch a GDB server with the specified command line,
     and launches GDB to attach to it.
@@ -280,6 +311,8 @@ def debug(args, gdbscript=None, exe=None, ssh=None, env=None, **kwargs):
         exe(str): Path to the executable on disk
         env(dict): Environment to start the binary in
         ssh(:class:`.ssh`): Remote ssh session to use to launch the process.
+        sysroot(str): Foreign-architecture sysroot, used for QEMU-emulated binaries
+            and Android targets.
 
     Returns:
         :class:`.process` or :class:`.ssh_channel`: A tube connected to the target process
@@ -290,14 +323,6 @@ def debug(args, gdbscript=None, exe=None, ssh=None, env=None, **kwargs):
         from the very beginning.  This requires that both ``gdb`` and ``gdbserver``
         are installed on your machine.
 
-        .. code-block:: python
-
-            # Create a new process, and stop it at 'main'
-            io = gdb.debug('bash', '''
-            break main
-            continue
-            ''')
-
         When GDB opens via :func:`debug`, it will initially be stopped on the very first
         instruction of the dynamic linker (``ld.so``) for dynamically-linked binaries.
 
@@ -322,74 +347,105 @@ def debug(args, gdbscript=None, exe=None, ssh=None, env=None, **kwargs):
             - Generally, you just add a few ``continue`` commands until things are set up
               the way you want it to be.
 
-        .. code-block:: python
+    Examples:
 
-            # Create a new process, and stop it at 'main'
-            io = gdb.debug('bash', '''
-            # Wait until we hit the main executable's entry point
-            break _start
-            continue
+    >>> # Create a new process, and stop it at 'main'
+    >>> io = gdb.debug('bash', '''
+    ... break main
+    ... continue
+    ... ''')
+    >>> io.sendline("echo hello")
+    >>> io.recvline()
+    b'hello\n'
+    >>> io.interactive() # doctest: +SKIP
+    >>> io.close()
 
-            # Now set breakpoint on shared library routines
-            break malloc
-            break free
-            continue
-            ''')
+    >>> # Create a new process, and stop it at '_start'
+    >>> io = gdb.debug('bash', '''
+    ... # Wait until we hit the main executable's entry point
+    ... break _start
+    ... continue
+    ...
+    ... # Now set breakpoint on shared library routines
+    ... break malloc
+    ... break free
+    ... continue
+    ... ''')
+    >>> io.sendline("echo hello")
+    >>> io.recvline()
+    b'hello\n'
+    >>> io.interactive() # doctest: +SKIP
+    >>> io.close()
 
-        You can use :func:`debug` to spawn new processes on remote machines as well,
-        by using the ``ssh=`` keyword to pass in your :class:`.ssh` instance.
+    You can use :func:`debug` to spawn new processes on remote machines as well,
+    by using the ``ssh=`` keyword to pass in your :class:`.ssh` instance.
 
-        .. code-block:: python
+    >>> # Connect to the SSH server
 
-            # Connect to the SSH server
-            shell = ssh('passcode', 'pwnable.kr', 2222, password='guest')
+    >>> # Start a process on the server
+    >>> shell = ssh('runner', 'example.pwnme', password='demopass')
+    >>> io = gdb.debug(['bash'],
+    ...                 ssh = shell,
+    ...                 gdbscript = '''
+    ... break main
+    ... continue
+    ... ''')
 
-            # Start a process on the server
-            bash = gdb.debug(['bash'],
-                            ssh=shell,
-                            gdbscript='''
-            break main
-            continue
-            ''')
+    >>> # Send a command to Bash
+    >>> io.sendline("echo hello")
+
+    >>> # Interact with the process
+    >>> io.interactive() # doctest: +SKIP
+    >>> io.close()
     """
     if isinstance(args, (int, tubes.process.process, tubes.ssh.ssh_channel)):
         log.error("Use gdb.attach() to debug a running process")
 
-    if env is None:
-        env = os.environ
-
-    if isinstance(args, (str, unicode)):
+    if isinstance(args, (bytes, six.text_type)):
         args = [args]
 
     orig_args = args
 
     runner = _get_runner(ssh)
     which  = _get_which(ssh)
+    gdbscript = gdbscript or ''
 
     if context.noptrace:
         log.warn_once("Skipping debugger since context.noptrace==True")
         return runner(args, executable=exe, env=env)
 
     if ssh or context.native or (context.os == 'android'):
-        args = _gdbserver_args(args=args, which=which)
+        args = _gdbserver_args(args=args, which=which, env=env)
     else:
         qemu_port = random.randint(1024, 65535)
-        args = [get_qemu_user(), '-g', str(qemu_port)] + args
+        qemu_user = qemu.user_path()
+        sysroot = sysroot or qemu.ld_prefix(env=env)
+        if not qemu_user:
+            log.error("Cannot debug %s binaries without appropriate QEMU binaries" % context.arch)
+        qemu_args = [qemu_user, '-g', str(qemu_port)]
+        if sysroot:
+            qemu_args += ['-L', sysroot]
+        args = qemu_args + args
+
+    # Use a sane default sysroot for Android
+    if not sysroot and context.os == 'android':
+        sysroot = 'remote:/'
 
     # Make sure gdbserver/qemu is installed
     if not which(args[0]):
         log.error("%s is not installed" % args[0])
 
-    exe = exe or which(orig_args[0])
-    if not exe:
-        log.error("%s does not exist" % orig_args[0])
+    if not ssh:
+        exe = exe or which(orig_args[0])
+        if not (exe and os.path.exists(exe)):
+            log.error("%s does not exist" % exe)
 
     # Start gdbserver/qemu
     # (Note: We override ASLR here for the gdbserver process itself.)
     gdbserver = runner(args, env=env, aslr=1, **kwargs)
 
     # Set the .executable on the process object.
-    gdbserver.executable = which(orig_args[0])
+    gdbserver.executable = exe
 
     # Find what port we need to connect to
     if context.native or (context.os == 'android'):
@@ -401,13 +457,13 @@ def debug(args, gdbscript=None, exe=None, ssh=None, env=None, **kwargs):
     if not ssh and context.os == 'android':
         host = context.adb_host
 
-    attach((host, port), exe=exe, gdbscript=gdbscript, need_ptrace_scope = False, ssh=ssh)
+    attach((host, port), exe=exe, gdbscript=gdbscript, need_ptrace_scope = False, ssh=ssh, sysroot=sysroot)
 
     # gdbserver outputs a message when a client connects
     garbage = gdbserver.recvline(timeout=1)
 
-    if "Remote debugging from host" not in garbage:
-        gdbserver.unrecv(garbage)
+    # Some versions of gdbserver output an additional message
+    garbage2 = gdbserver.recvline_startswith(b"Remote debugging from host ", timeout=1)
 
     return gdbserver
 
@@ -417,7 +473,8 @@ def get_gdb_arch():
         'powerpc': 'powerpc:common',
         'powerpc64': 'powerpc:common64',
         'mips64': 'mips:isa64',
-        'thumb': 'arm'
+        'thumb': 'arm',
+        'sparc64': 'sparc:v9'
     }.get(context.arch, context.arch)
 
 def binary():
@@ -425,8 +482,13 @@ def binary():
 
     Returns:
         str: Path to the appropriate ``gdb`` binary to use.
+
+    Example:
+
+        >>> gdb.binary() # doctest: +SKIP
+        '/usr/bin/gdb'
     """
-    gdb = misc.which('gdb')
+    gdb = misc.which('pwntools-gdb') or misc.which('gdb')
 
     if not context.native:
         multiarch = misc.which('gdb-multiarch')
@@ -443,8 +505,8 @@ def binary():
     return gdb
 
 @LocalContext
-def attach(target, gdbscript = None, exe = None, need_ptrace_scope = True, gdb_args = None, ssh = None):
-    """attach(target, gdbscript = None, exe = None, arch = None, ssh = None) -> None
+def attach(target, gdbscript = '', exe = None, need_ptrace_scope = True, gdb_args = None, ssh = None, sysroot = None):
+    r"""attach(target, gdbscript = None, exe = None, arch = None, ssh = None) -> None
 
     Start GDB in a new terminal and attach to `target`.
 
@@ -455,6 +517,8 @@ def attach(target, gdbscript = None, exe = None, need_ptrace_scope = True, gdb_a
         arch(str): Architechture of the target binary.  If `exe` known GDB will
           detect the architechture automatically (if it is supported).
         gdb_args(list): List of additional arguments to pass to GDB.
+        sysroot(str): Foreign-architecture sysroot, used for QEMU-emulated binaries
+            and Android targets.
 
     Returns:
         PID of the GDB process (or the window which it is running in).
@@ -479,64 +543,70 @@ def attach(target, gdbscript = None, exe = None, need_ptrace_scope = True, gdb_a
             This will use the GDB installed on the remote machine.
             If a password is required to connect, the ``sshpass`` program must be installed.
 
-        .. code-block:: python
+    Examples:
 
-            # Attach directly to pid 1234
-            gdb.attach(1234)
+    >>> # Attach directly to pid 1234
+    >>> gdb.attach(1234) # doctest: +SKIP
 
-        .. code-block:: python
 
-            # Attach to the youngest "bash" process
-            gdb.attach('bash')
+    >>> # Attach to the youngest "bash" process
+    >>> gdb.attach('bash') # doctest: +SKIP
 
-        .. code-block:: python
 
-            # Start a process
-            bash = process('bash')
+    >>> # Start a process
+    >>> bash = process('bash')
 
-            # Attach the debugger
-            gdb.attach(bash, '''
-            set follow-fork-mode child
-            break execve
-            continue
-            ''')
+    >>> # Attach the debugger
+    >>> pid = gdb.attach(bash, '''
+    ... set follow-fork-mode child
+    ... break execve
+    ... continue
+    ... ''')
 
-            # Interact with the process
-            bash.sendline('whoami')
+    >>> # Interact with the process
+    >>> bash.sendline("whoami")
+    >>> bash.recvline()
+    b'runner\n'
+    >>> bash.close()
 
-        .. code-block:: python
+    >>> # Start a forking server
+    >>> server = process(['socat', 'tcp-listen:12345,fork,reuseaddr', 'exec:/bin/bash,nofork'])
+    >>> sleep(1)
 
-            # Start a forking server
-            server = process(['socat', 'tcp-listen:1234,fork,reuseaddr', 'exec:/bin/sh'])
+    >>> # Connect to the server
+    >>> io = remote('127.0.0.1', 12345)
 
-            # Connect to the server
-            io = remote('localhost', 1234)
+    >>> # Connect the debugger to the server-spawned process
+    >>> pid = gdb.attach(io, '''
+    ... break exit
+    ... continue
+    ... ''', exe = '/bin/bash')
 
-            # Connect the debugger to the server-spawned process
-            gdb.attach(io, '''
-            break exit
-            continue
-            ''')
+    >>> # Talk to the spawned 'sh'
+    >>> io.sendline("echo hello")
+    >>> io.recvline()
+    b'hello\n'
+    >>> io.sendline("exit")
 
-            # Talk to the spawned 'sh'
-            io.sendline('exit')
+    >>> io.close()
 
-        .. code-block:: python
+    >>> # Connect to the SSH server
+    >>> shell = ssh('runner', 'example.pwnme', password='demopass')
 
-            # Connect to the SSH server
-            shell = ssh('bandit0', 'bandit.labs.overthewire.org', password='bandit0', port=2220)
+    >>> # Start a process on the server
+    >>> cat = shell.process(['cat'])
 
-            # Start a process on the server
-            cat = shell.process(['cat'])
+    >>> # Attach a debugger to it
+    >>> gdb.attach(cat, '''
+    ... break exit
+    ... continue
+    ... ''')
 
-            # Attach a debugger to it
-            gdb.attach(cat, '''
-            break exit
-            continue
-            ''')
-
-            # Cause `cat` to exit
-            cat.close()
+    >>> cat.sendline("hello")
+    >>> cat.recvline()
+    b'hello\n'
+    >>> # Cause `cat` to exit
+    >>> cat.close()
     """
     if context.noptrace:
         log.warn_once("Skipping debug attach since context.noptrace==True")
@@ -544,7 +614,7 @@ def attach(target, gdbscript = None, exe = None, need_ptrace_scope = True, gdb_a
 
     # if gdbscript is a file object, then read it; we probably need to run some
     # more gdb script anyway
-    if isinstance(gdbscript, file):
+    if hasattr(gdbscript, 'read'):
         with gdbscript:
             gdbscript = gdbscript.read()
 
@@ -552,18 +622,24 @@ def attach(target, gdbscript = None, exe = None, need_ptrace_scope = True, gdb_a
     if gdbscript and not gdbscript.endswith('\n'):
         gdbscript += '\n'
 
+    # Use a sane default sysroot for Android
+    if not sysroot and context.os == 'android':
+        sysroot = 'remote:/'
+
     # gdb script to run before `gdbscript`
     pre = ''
     if not context.native:
         pre += 'set endian %s\n' % context.endian
         pre += 'set architecture %s\n' % get_gdb_arch()
+        if sysroot:
+            pre += 'set sysroot %s\n' % sysroot
 
         if context.os == 'android':
             pre += 'set gnutarget ' + _bfdname() + '\n'
 
     # let's see if we can find a pid to attach to
     pid = None
-    if   isinstance(target, (int, long)):
+    if   isinstance(target, six.integer_types):
         # target is a pid, easy peasy
         pid = target
     elif isinstance(target, str):
@@ -586,6 +662,8 @@ def attach(target, gdbscript = None, exe = None, need_ptrace_scope = True, gdb_a
         shell = target.parent
 
         tmpfile = shell.mktemp()
+        if six.PY3:
+            tmpfile = tmpfile.decode()
         gdbscript = 'shell rm %s\n%s' % (tmpfile, gdbscript)
         shell.upload_data(gdbscript or '', tmpfile)
 
@@ -596,7 +674,10 @@ def attach(target, gdbscript = None, exe = None, need_ptrace_scope = True, gdb_a
             cmd = ['sshpass', '-p', shell.password] + cmd
         if shell.keyfile:
             cmd += ['-i', shell.keyfile]
-        cmd += ['gdb -q %r %s -x "%s"' % (target.executable,
+        exefile = target.executable
+        if six.PY3:
+            exefile = exefile.decode()
+        cmd += ['gdb -q %s %s -x "%s"' % (exefile,
                                        target.pid,
                                        tmpfile)]
 
@@ -608,12 +689,29 @@ def attach(target, gdbscript = None, exe = None, need_ptrace_scope = True, gdb_a
         if not pids:
             log.error('could not find remote process (%s:%d) on this machine' %
                       target.sock.getpeername())
-        pid = pids[0]
+        waiting = True
+        if exe:
+            while waiting:
+                for pid in pids:
+                    if proc.exe(pid) == exe:
+                        waiting = False
+                        break
+                else:
+                    time.sleep(0.01)
     elif isinstance(target, tubes.process.process):
         pid = proc.pidof(target)[0]
+        exe = exe or target.executable
     elif isinstance(target, tuple) and len(target) == 2:
         host, port = target
-        pre += 'target remote %s:%d\n' % (host, port)
+
+        if context.os != 'android':
+            pre += 'target remote %s:%d\n' % (host, port)
+        else:
+            # Android debugging is done over gdbserver, which can't follow
+            # new inferiors (tldr; follow-fork-mode child) unless it is run
+            # in extended-remote mode.
+            pre += 'target extended-remote %s:%d\n' % (host, port)
+            pre += 'set detach-on-fork off\n'
 
         def findexe():
             for spid in proc.pidof(target):
@@ -637,7 +735,7 @@ def attach(target, gdbscript = None, exe = None, need_ptrace_scope = True, gdb_a
             exe_fn = adb.proc_exe
         exe = exe_fn(pid)
 
-    if not pid and not exe:
+    if not pid and not exe and not ssh:
         log.error('could not find target process')
 
     cmd = binary()
@@ -646,13 +744,14 @@ def attach(target, gdbscript = None, exe = None, need_ptrace_scope = True, gdb_a
         cmd += ' '
         cmd += ' '.join(gdb_args)
 
+    if context.gdbinit:
+        cmd += ' -nh '                     # ignore ~/.gdbinit
+        cmd += ' -x %s ' % context.gdbinit # load custom gdbinit
+
     cmd += ' -q '
 
     if exe and context.native:
-        if ssh:
-            ssh.download_file(exe)
-            exe = os.path.basename(exe)
-        if not os.path.isfile(exe):
+        if not ssh and not os.path.isfile(exe):
             log.error('No such file: %s' % exe)
         cmd += ' "%s"' % exe
 
@@ -662,17 +761,21 @@ def attach(target, gdbscript = None, exe = None, need_ptrace_scope = True, gdb_a
     if context.os == 'android' and pid:
         runner  = _get_runner()
         which   = _get_which()
-        gdb_cmd = _gdbserver_args(pid=pid, which=which)
+        gdb_cmd = _gdbserver_args(pid=pid, which=which, env=env)
         gdbserver = runner(gdb_cmd)
         port    = _gdbserver_port(gdbserver, None)
         host    = context.adb_host
-        pre    += 'target remote %s:%i' % (context.adb_host, port)
+        pre    += 'target extended-remote %s:%i\n' % (context.adb_host, port)
+
+        # gdbserver on Android sets 'detach-on-fork on' which breaks things
+        # when you're trying to debug anything that forks.
+        pre += 'set detach-on-fork off\n'
 
     gdbscript = pre + (gdbscript or '')
 
     if gdbscript:
         tmp = tempfile.NamedTemporaryFile(prefix = 'pwn', suffix = '.gdb',
-                                          delete = False)
+                                          delete = False, mode = 'w+')
         log.debug('Wrote gdb script to %r\n%s' % (tmp.name, gdbscript))
         gdbscript = 'shell rm %s\n%s' % (tmp.name, gdbscript)
 
@@ -685,7 +788,7 @@ def attach(target, gdbscript = None, exe = None, need_ptrace_scope = True, gdb_a
     gdb_pid = misc.run_in_new_terminal(cmd)
 
     if pid and context.native:
-        proc.wait_for_debugger(pid)
+        proc.wait_for_debugger(pid, gdb_pid)
 
     return gdb_pid
 
@@ -694,7 +797,7 @@ def ssh_gdb(ssh, argv, gdbscript = None, arch = None, **kwargs):
         argv = [argv]
 
     exe = argv[0]
-    argv = ["gdbserver", "127.0.0.1:0"] + argv
+    argv = ["gdbserver", "--multi", "127.0.0.1:0"] + argv
 
     # Download the executable
     local_exe = os.path.basename(exe)
@@ -704,9 +807,9 @@ def ssh_gdb(ssh, argv, gdbscript = None, arch = None, **kwargs):
     c = ssh.process(argv, **kwargs)
 
     # Find the port for the gdb server
-    c.recvuntil('port ')
+    c.recvuntil(b'port ')
     line = c.recvline().strip()
-    gdbport = re.match('[0-9]+', line)
+    gdbport = re.match(b'[0-9]+', line)
     if gdbport:
         gdbport = int(gdbport.group(0))
 
@@ -714,7 +817,7 @@ def ssh_gdb(ssh, argv, gdbscript = None, arch = None, **kwargs):
     forwardport = l.lport
 
     attach(('127.0.0.1', forwardport), gdbscript, local_exe, arch, ssh=ssh)
-    l.wait_for_connection() <> ssh.connect_remote('127.0.0.1', gdbport)
+    l.wait_for_connection().connect_both(ssh.connect_remote('127.0.0.1', gdbport))
     return c
 
 def find_module_addresses(binary, ssh=None, ulimit=False):
@@ -762,10 +865,10 @@ def find_module_addresses(binary, ssh=None, ulimit=False):
 
     Example:
 
-    >>> with context.local(log_level=9999): # doctest: +SKIP
-    ...     shell = ssh(host='bandit.labs.overthewire.org',user='bandit0',password='bandit0', port=2220)
+    >>> with context.local(log_level=9999):
+    ...     shell =  ssh(host='example.pwnme', user='runner', password='demopass')
     ...     bash_libs = gdb.find_module_addresses('/bin/bash', shell)
-    >>> os.path.basename(bash_libs[0].path) # doctest: +SKIP
+    >>> os.path.basename(bash_libs[0].path)
     'libc.so.6'
     >>> hex(bash_libs[0].symbols['system']) # doctest: +SKIP
     '0x7ffff7634660'
@@ -784,31 +887,29 @@ def find_module_addresses(binary, ssh=None, ulimit=False):
         local_elf  = elf.ELF(binary)
         local_libs = local_elf.libs
 
-    entry      = local_elf.header.e_entry
-
     #
     # Get the addresses from GDB
     #
     libs = {}
-    cmd  = "gdb -q --args %s" % (binary)
+    cmd  = "gdb -q -nh --args %s | cat" % (binary) # pipe through cat to disable colored output on GDB 9+
     expr = re.compile(r'(0x\S+)[^/]+(.*)')
 
     if ulimit:
-        cmd = 'sh -c "(ulimit -s unlimited; %s)"' % cmd
-
-    cmd = shlex.split(cmd)
+        cmd = ['sh', '-c', "(ulimit -s unlimited; %s)" % cmd]
+    else:
+        cmd = ['sh', '-c', cmd]
 
     with runner(cmd) as gdb:
         if context.aslr:
             gdb.sendline('set disable-randomization off')
+
         gdb.send("""
         set prompt
-        break *%#x
+        catch load
         run
-        """ % entry)
-        gdb.clean(2)
+        """)
         gdb.sendline('info sharedlibrary')
-        lines = gdb.recvrepeat(2)
+        lines = context._decode(gdb.recvrepeat(2))
 
         for line in lines.splitlines():
             m = expr.match(line)
@@ -828,7 +929,7 @@ def find_module_addresses(binary, ssh=None, ulimit=False):
         try:
             path     = next(p for p in local_libs.keys() if remote_path in p)
         except StopIteration:
-            print "Skipping %r" % remote_path
+            print("Skipping %r" % remote_path)
             continue
 
         # Load it
@@ -898,15 +999,15 @@ def version(program='gdb'):
 
     Example:
 
-        >>> (7,0) <= gdb.version() <= (8,0)
+        >>> (7,0) <= gdb.version() <= (10,0)
         True
     """
     program = misc.which(program)
-    expr = r'([0-9]+\.?)+'
+    expr = br'([0-9]+\.?)+'
 
     with tubes.process.process([program, '--version'], level='error') as gdb:
         version = gdb.recvline()
 
     versions = re.search(expr, version).group()
 
-    return tuple(map(int, versions.split('.')))
+    return tuple(map(int, versions.split(b'.')))
