@@ -10,7 +10,7 @@ have been useful.
 
     >>> context.clear(arch='i386')
     >>> binary = ELF.from_assembly('add esp, 0x10; ret')
-    >>> binary.symbols = {'read': 0xdeadbeef, 'write': 0xdecafbad, 'exit': 0xfeedface}
+    >>> binary.symbols = {'read': 0xdeadbeef, 'write': 0xdecafbad, 'execve': 0xcafebabe, 'exit': 0xfeedface}
 
 Creating a ROP object which looks up symbols in the binary is
 pretty straightforward.
@@ -20,13 +20,13 @@ pretty straightforward.
 With the ROP object, you can manually add stack frames.
 
     >>> rop.raw(0)
-    >>> rop.raw(unpack('abcd'))
+    >>> rop.raw(unpack(b'abcd'))
     >>> rop.raw(2)
 
 Inspecting the ROP stack is easy, and laid out in an easy-to-read
 manner.
 
-    >>> print rop.dump()
+    >>> print(rop.dump())
     0x0000:              0x0
     0x0004:       0x64636261
     0x0008:              0x2
@@ -35,12 +35,12 @@ The ROP module is also aware of how to make function calls with
 standard Linux ABIs.
 
     >>> rop.call('read', [4,5,6])
-    >>> print rop.dump()
+    >>> print(rop.dump())
     0x0000:              0x0
     0x0004:       0x64636261
     0x0008:              0x2
     0x000c:       0xdeadbeef read(4, 5, 6)
-    0x0010:           'eaaa' <pad>
+    0x0010:          b'eaaa' <return address>
     0x0014:              0x4 arg0
     0x0018:              0x5 arg1
     0x001c:              0x6 arg2
@@ -50,24 +50,44 @@ The stack is automatically adjusted for the next frame
 
     >>> rop.write(7,8,9)
     >>> rop.exit()
-    >>> print rop.dump()
+    >>> print(rop.dump())
     0x0000:              0x0
     0x0004:       0x64636261
     0x0008:              0x2
     0x000c:       0xdeadbeef read(4, 5, 6)
-    0x0010:       0x10000000 <adjust: add esp, 0x10; ret>
+    0x0010:       0x10000000 <adjust @0x24> add esp, 0x10; ret
     0x0014:              0x4 arg0
     0x0018:              0x5 arg1
     0x001c:              0x6 arg2
-    0x0020:           'iaaa' <pad>
+    0x0020:          b'iaaa' <pad>
     0x0024:       0xdecafbad write(7, 8, 9)
-    0x0028:       0x10000000 <adjust: add esp, 0x10; ret>
+    0x0028:       0x10000000 <adjust @0x3c> add esp, 0x10; ret
     0x002c:              0x7 arg0
     0x0030:              0x8 arg1
     0x0034:              0x9 arg2
-    0x0038:           'oaaa' <pad>
+    0x0038:          b'oaaa' <pad>
     0x003c:       0xfeedface exit()
-    0x0040:           'qaaa' <pad>
+
+You can also append complex arguments onto stack when the stack pointer is known.
+
+    >>> rop = ROP(binary, base=0x7fffe000)
+    >>> rop.call('execve', ['/bin/sh', [['/bin/sh'], ['-p'], ['-c'], ['ls']], 0])
+    >>> print(rop.dump())
+    0x7fffe000:       0xcafebabe execve(['/bin/sh'], [['/bin/sh'], ['-p'], ['-c'], ['ls']], 0)
+    0x7fffe004:          b'baaa' <return address>
+    0x7fffe008:       0x7fffe014 arg0 (+0xc)
+    0x7fffe00c:       0x7fffe01c arg1 (+0x10)
+    0x7fffe010:              0x0 arg2
+    0x7fffe014:   b'/bin/sh\x00'
+    0x7fffe01c:       0x7fffe02c (+0x10)
+    0x7fffe020:       0x7fffe034 (+0x14)
+    0x7fffe024:       0x7fffe038 (+0x14)
+    0x7fffe028:       0x7fffe03c (+0x14)
+    0x7fffe02c:   b'/bin/sh\x00'
+    0x7fffe034:       b'-p\x00$'
+    0x7fffe038:       b'-c\x00$'
+    0x7fffe03c:       b'ls\x00$'
+
 
 ROP Example
 -------------------
@@ -105,28 +125,84 @@ Finally, let's build our ROP stack
     >>> rop = ROP(binary)
     >>> rop.write(c.STDOUT_FILENO, binary.symbols['flag'], 8)
     >>> rop.exit()
-    >>> print rop.dump()
-    0x0000:       0x10000012 write(STDOUT_FILENO, 268435494, 8)
-    0x0004:       0x1000000e <adjust: add esp, 0x10; ret>
-    0x0008:              0x1 arg0
+    >>> print(rop.dump())
+    0x0000:       0x10000012 write(STDOUT_FILENO, 0x10000026, 8)
+    0x0004:       0x1000000e <adjust @0x18> add esp, 0x10; ret
+    0x0008:              0x1 STDOUT_FILENO
     0x000c:       0x10000026 flag
     0x0010:              0x8 arg2
-    0x0014:           'faaa' <pad>
+    0x0014:          b'faaa' <pad>
     0x0018:       0x1000002f exit()
-    0x001c:           'haaa' <pad>
 
 The raw data from the ROP stack is available via `str`.
 
-    >>> raw_rop = str(rop)
-    >>> print enhex(raw_rop)
-    120000100e000010010000002600001008000000666161612f00001068616161
+    >>> raw_rop = rop.chain()
+    >>> print(enhex(raw_rop))
+    120000100e000010010000002600001008000000666161612f000010
 
 Let's try it out!
 
     >>> p = process(binary.path)
     >>> p.send(raw_rop)
-    >>> print p.recvall(timeout=5)
-    The flag
+    >>> print(repr(p.recvall(timeout=5)))
+    b'The flag'
+
+ROP Example (amd64)
+-------------------
+
+For amd64 binaries, the registers are loaded off the stack.  Pwntools can do
+basic reasoning about simple "pop; pop; add; ret"-style gadgets, and satisfy
+requirements so that everything "just works".
+
+    >>> context.clear(arch='amd64')
+    >>> assembly = 'pop rdx; pop rdi; pop rsi; add rsp, 0x20; ret; target: ret'
+    >>> binary = ELF.from_assembly(assembly)
+    >>> rop = ROP(binary)
+    >>> rop.target(1,2,3)
+    >>> print(rop.dump())
+    0x0000:       0x10000000 pop rdx; pop rdi; pop rsi; add rsp, 0x20; ret
+    0x0008:              0x3 [arg2] rdx = 3
+    0x0010:              0x1 [arg0] rdi = 1
+    0x0018:              0x2 [arg1] rsi = 2
+    0x0020:      b'iaaajaaa' <pad 0x20>
+    0x0028:      b'kaaalaaa' <pad 0x18>
+    0x0030:      b'maaanaaa' <pad 0x10>
+    0x0038:      b'oaaapaaa' <pad 0x8>
+    0x0040:       0x10000008 target
+    >>> rop.target(1)
+    >>> print(rop.dump())
+    0x0000:       0x10000000 pop rdx; pop rdi; pop rsi; add rsp, 0x20; ret
+    0x0008:              0x3 [arg2] rdx = 3
+    0x0010:              0x1 [arg0] rdi = 1
+    0x0018:              0x2 [arg1] rsi = 2
+    0x0020:      b'iaaajaaa' <pad 0x20>
+    0x0028:      b'kaaalaaa' <pad 0x18>
+    0x0030:      b'maaanaaa' <pad 0x10>
+    0x0038:      b'oaaapaaa' <pad 0x8>
+    0x0040:       0x10000008 target
+    0x0048:       0x10000001 pop rdi; pop rsi; add rsp, 0x20; ret
+    0x0050:              0x1 [arg0] rdi = 1
+    0x0058:      b'waaaxaaa' <pad rsi>
+    0x0060:      b'yaaazaab' <pad 0x20>
+    0x0068:      b'baabcaab' <pad 0x18>
+    0x0070:      b'daabeaab' <pad 0x10>
+    0x0078:      b'faabgaab' <pad 0x8>
+    0x0080:       0x10000008 target
+
+Pwntools will also filter out some bad instructions while setting the registers
+( e.g. syscall, int 0x80... )
+
+    >>> assembly = 'syscall; pop rdx; pop rsi; ret ; pop rdi ; int 0x80; pop rsi; pop rdx; ret ; pop rdi ; ret'
+    >>> binary = ELF.from_assembly(assembly)
+    >>> rop = ROP(binary)
+    >>> rop.call(0xdeadbeef, [1, 2, 3])
+    >>> print(rop.dump())
+    0x0000:       0x1000000b pop rdi; ret
+    0x0008:              0x1 [arg0] rdi = 1
+    0x0010:       0x10000002 pop rdx; pop rsi; ret
+    0x0018:              0x3 [arg2] rdx = 3
+    0x0020:              0x2 [arg1] rsi = 2
+    0x0028:       0xdeadbeef
 
 ROP + Sigreturn
 -----------------------
@@ -158,9 +234,9 @@ Let's create a ROP object and invoke the call.
 
 That's all there is to it.
 
-    >>> print rop.dump()
+    >>> print(rop.dump())
     0x0000:       0x1000000e pop eax; ret
-    0x0004:             0x77
+    0x0004:             0x77 [arg0] eax = SYS_sigreturn
     0x0008:       0x1000000b int 0x80
     0x000c:              0x0 gs
     0x0010:              0x0 fs
@@ -173,7 +249,7 @@ That's all there is to it.
     0x002c:       0x10000012 ebx = binsh
     0x0030:              0x0 edx
     0x0034:              0x0 ecx
-    0x0038:              0xb eax
+    0x0038:              0xb eax = SYS_execve
     0x003c:              0x0 trapno
     0x0040:              0x0 err
     0x0044:       0x1000000b int 0x80
@@ -186,19 +262,24 @@ That's all there is to it.
 Let's try it out!
 
     >>> p = process(binary.path)
-    >>> p.send(str(rop))
+    >>> p.send(rop.chain())
     >>> time.sleep(1)
-    >>> p.sendline('echo hello; exit')
+    >>> p.sendline(b'echo hello; exit')
     >>> p.recvline()
-    'hello\n'
+    b'hello\n'
 """
 from __future__ import absolute_import
+from __future__ import division
 
 import collections
 import copy
 import hashlib
+import itertools
 import os
 import re
+import shutil
+import six
+import string
 import sys
 import tempfile
 
@@ -214,11 +295,12 @@ from pwnlib.rop.call import Call
 from pwnlib.rop.call import CurrentStackPointer
 from pwnlib.rop.call import NextGadgetAddress
 from pwnlib.rop.call import StackAdjustment
+from pwnlib.rop.call import Unresolved
 from pwnlib.rop.gadgets import Gadget
-from pwnlib.util import cyclic
 from pwnlib.util import lists
 from pwnlib.util import packing
-from pwnlib.util.packing import *
+from pwnlib.util.cyclic import cyclic
+from pwnlib.util.packing import pack
 
 log = getLogger(__name__)
 __all__ = ['ROP']
@@ -228,26 +310,39 @@ class Padding(object):
     """
     Placeholder for exactly one pointer-width of padding.
     """
+    def __init__(self, name='<pad>'):
+        self.name = name
+
+def _slot_len(x):
+    if isinstance(x, six.integer_types+(Unresolved, Padding, Gadget)):
+        return context.bytes
+    else:
+        return len(packing.flat(x))
 
 class DescriptiveStack(list):
     """
     List of resolved ROP gadgets that correspond to the ROP calls that
-    the user has specified.  Also includes
+    the user has specified.
     """
 
     #: Base address
     address = 0
 
-    #: Dictionary of ``{address: [list of descriptions]}``
+    #: Dictionary of \`{address: [list of descriptions]}`
     descriptions = {}
 
     def __init__(self, address):
         self.descriptions = collections.defaultdict(lambda: [])
         self.address      = address or 0
+        self._next_next   = 0
+        self._next_last   = 0
 
     @property
     def next(self):
-        return self.address + len(self) * context.bytes
+        for x in self[self._next_last:]:
+            self._next_next += _slot_len(x)
+        self._next_last = len(self)
+        return self.address + self._next_next
 
     def describe(self, text, address = None):
         if address is None:
@@ -256,13 +351,13 @@ class DescriptiveStack(list):
 
     def dump(self):
         rv = []
+        addr = self.address
         for i, data in enumerate(self):
-            addr = self.address + i * context.bytes
             off = None
             line = '0x%04x:' % addr
-            if isinstance(data, str):
+            if isinstance(data, (str, bytes)):
                 line += ' %16r' % data
-            elif isinstance(data, (int,long)):
+            elif isinstance(data, six.integer_types):
                 line += ' %#16x' % data
                 if self.address != 0 and self.address < data < self.next:
                     off = data - addr
@@ -274,6 +369,7 @@ class DescriptiveStack(list):
             if off is not None:
                 line += ' (+%#x)' % off
             rv.append(line)
+            addr += _slot_len(data)
 
         return '\n'.join(rv)
 
@@ -293,7 +389,7 @@ class ROP(object):
        #  '0x0004:       0xdeadbeef',
        #  '0x0008:              0x0',
        #  '0x000c:        0x80496a8']
-       str(rop)
+       bytes(rop)
        # '\xfc\x82\x04\x08\xef\xbe\xad\xde\x00\x00\x00\x00\xa8\x96\x04\x08'
 
     >>> context.clear(arch = "i386", kernel = 'amd64')
@@ -304,18 +400,18 @@ class ROP(object):
     >>> r.funcname(1, 2)
     >>> r.funcname(3)
     >>> r.execve(4, 5, 6)
-    >>> print r.dump()
+    >>> print(r.dump())
     0x0000:       0x10001234 funcname(1, 2)
-    0x0004:       0x10000003 <adjust: add esp, 0x10; ret>
+    0x0004:       0x10000003 <adjust @0x18> add esp, 0x10; ret
     0x0008:              0x1 arg0
     0x000c:              0x2 arg1
-    0x0010:           'eaaa' <pad>
-    0x0014:           'faaa' <pad>
+    0x0010:          b'eaaa' <pad>
+    0x0014:          b'faaa' <pad>
     0x0018:       0x10001234 funcname(3)
-    0x001c:       0x10000007 <adjust: pop eax; ret>
+    0x001c:       0x10000007 <adjust @0x24> pop eax; ret
     0x0020:              0x3 arg0
     0x0024:       0x10000007 pop eax; ret
-    0x0028:             0x77
+    0x0028:             0x77 [arg0] eax = SYS_sigreturn
     0x002c:       0x10000000 int 0x80
     0x0030:              0x0 gs
     0x0034:              0x0 fs
@@ -328,7 +424,7 @@ class ROP(object):
     0x0050:              0x4 ebx
     0x0054:              0x6 edx
     0x0058:              0x5 ecx
-    0x005c:              0xb eax
+    0x005c:              0xb eax = SYS_execve
     0x0060:              0x0 trapno
     0x0064:              0x0 err
     0x0068:       0x10000000 int 0x80
@@ -342,18 +438,18 @@ class ROP(object):
     >>> r.funcname(1, 2)
     >>> r.funcname(3)
     >>> r.execve(4, 5, 6)
-    >>> print r.dump()
+    >>> print(r.dump())
     0x8048000:       0x10001234 funcname(1, 2)
-    0x8048004:       0x10000003 <adjust: add esp, 0x10; ret>
+    0x8048004:       0x10000003 <adjust @0x8048018> add esp, 0x10; ret
     0x8048008:              0x1 arg0
     0x804800c:              0x2 arg1
-    0x8048010:           'eaaa' <pad>
-    0x8048014:           'faaa' <pad>
+    0x8048010:          b'eaaa' <pad>
+    0x8048014:          b'faaa' <pad>
     0x8048018:       0x10001234 funcname(3)
-    0x804801c:       0x10000007 <adjust: pop eax; ret>
+    0x804801c:       0x10000007 <adjust @0x8048024> pop eax; ret
     0x8048020:              0x3 arg0
     0x8048024:       0x10000007 pop eax; ret
-    0x8048028:             0x77
+    0x8048028:             0x77 [arg0] eax = SYS_sigreturn
     0x804802c:       0x10000000 int 0x80
     0x8048030:              0x0 gs
     0x8048034:              0x0 fs
@@ -366,7 +462,7 @@ class ROP(object):
     0x8048050:              0x4 ebx
     0x8048054:              0x6 edx
     0x8048058:              0x5 ecx
-    0x804805c:              0xb eax
+    0x804805c:              0xb eax = SYS_execve
     0x8048060:              0x0 trapno
     0x8048064:              0x0 err
     0x8048068:       0x10000000 int 0x80
@@ -375,37 +471,50 @@ class ROP(object):
     0x8048074:              0x0 esp_at_signal
     0x8048078:             0x2b ss
     0x804807c:              0x0 fpstate
+
+
+    >>> elf = ELF.from_assembly('ret')
+    >>> r = ROP(elf)
+    >>> r.ret.address == 0x10000000
+    True
+    >>> r = ROP(elf, badchars=b'\x00')
+    >>> r.gadgets == {}
+    True
+    >>> r.ret is None
+    True
     """
-    #: List of individual ROP gadgets, ROP calls, SROP frames, etc.
-    #: This is intended to be the highest-level abstraction that we can muster.
-    _chain = []
-
-    #: List of ELF files which are available for mining gadgets
-    elfs = []
-
-    #: Stack address where the first byte of the ROP chain lies, if known.
-    base = 0
-
-    #: Whether or not the ROP chain directly sets the stack pointer to a value
-    #: which is not contiguous
-    migrated = False
-
-    def __init__(self, elfs, base = None, **kwargs):
+    def __init__(self, elfs, base = None, badchars = b'', **kwargs):
         """
         Arguments:
             elfs(list): List of :class:`.ELF` objects for mining
+            base(int): Stack address where the first byte of the ROP chain lies, if known.
+            badchars(str): Characters which should not appear in ROP gadget addresses.
         """
         import ropgadget
 
         # Permit singular ROP(elf) vs ROP([elf])
         if isinstance(elfs, ELF):
             elfs = [elfs]
-        elif isinstance(elfs, (str, unicode)):
+        elif isinstance(elfs, (bytes, six.text_type)):
             elfs = [ELF(elfs)]
-        self.elfs = elfs
+
+        #: List of individual ROP gadgets, ROP calls, SROP frames, etc.
+        #: This is intended to be the highest-level abstraction that we can muster.
         self._chain = []
+
+        #: List of ELF files which are available for mining gadgets
+        self.elfs = elfs
+
+        #: Stack address where the first byte of the ROP chain lies, if known.
         self.base = base
+
+        #: Whether or not the ROP chain directly sets the stack pointer to a value
+        #: which is not contiguous
         self.migrated = False
+
+        #: Characters which should not appear in ROP gadget addresses.
+        self._badchars = set(badchars)
+
         self.__load()
 
     @staticmethod
@@ -415,24 +524,102 @@ class ROP(object):
 
     def setRegisters(self, registers):
         """
-        Returns an OrderedDict of addresses/values which will set the specified
-        register context.
+        Returns an list of addresses/values which will set the specified register context.
 
         Arguments:
             registers(dict): Dictionary of ``{register name: value}``
 
         Returns:
-            An OrderedDict of ``{register: sequence of gadgets, values, etc.}``.
+            A list of tuples, ordering the stack.
+
+            Each tuple is in the form of ``(value, name)`` where ``value`` is either a
+            gadget address or literal value to go on the stack, and ``name`` is either
+            a string name or other item which can be "unresolved".
+
+        Note:
+            This is basically an implementation of the Set Cover Problem, which is
+            NP-hard.  This means that we will take polynomial time N**2, where N is
+            the number of gadgets.  We can reduce runtime by discarding useless and
+            inferior gadgets ahead of time.
         """
-        reg_order = collections.OrderedDict()
+        if not registers:
+            return []
 
-        for reg, value in registers.items():
-            gadget = self.find_gadget(['pop ' + reg, 'ret'])
-            if not gadget:
-                log.error("can't set %r" % reg)
-            reg_order[reg] = [gadget, value]
+        regset = set(registers)
 
-        return reg_order
+        bad_instructions = set(('syscall', 'sysenter', 'int 0x80'))
+        
+        # Collect all gadgets which use these registers
+        # Also collect the "best" gadget for each combination of registers
+        gadgets = []
+        best_gadgets = {}
+
+        for gadget in self.gadgets.values():
+            # Do not use gadgets which doesn't end with 'ret'
+            if gadget.insns[-1] != 'ret':
+                continue
+            # Do not use gadgets which contain 'syscall' or 'int'
+            if set(gadget.insns) & bad_instructions:
+                continue
+
+            touched = tuple(regset & set(gadget.regs))
+
+            if not touched:
+                continue
+
+            old = best_gadgets.get(touched, gadget)
+
+            # if we have a new gadget for the touched registers, choose it
+            # if the new gadget requires less stack space, choose it
+            # if both gadgets require same stack space, choose the one with less instructions
+            if (old is gadget) \
+              or (old.move > gadget.move) \
+              or (old.move == gadget.move and len(old.insns) > len(gadget.insns)):
+                best_gadgets[touched] = gadget
+
+        winner = None
+        budget = 999999999
+
+        for num_gadgets in range(len(registers)):
+            for combo in itertools.combinations(sorted(best_gadgets.values(), key=repr, reverse=True), 1+num_gadgets):
+                # Is this better than what we can already do?
+                cost = sum((g.move for g in combo))
+                if cost > budget:
+                    continue
+
+                # Does it hit all of the registers we want?
+                coverage = set(sum((g.regs for g in combo), [])) & regset
+
+                if coverage != regset:
+                    continue
+
+                # It is better than what we had, and hits all of the registers.
+                winner = combo
+                budget = cost
+
+        if not winner:
+            log.error("Could not satisfy setRegisters(%r)", registers)
+
+        # We have our set of "winner" gadgets, let's build a stack!
+        stack = []
+
+        for gadget in winner:
+            moved = 8 # Account for the gadget itself
+            goodregs = set(gadget.regs) & regset
+            name = ",".join(goodregs)
+            stack.append((gadget.address, gadget))
+            for r in gadget.regs:
+                moved += 8
+                if r in registers:
+                    stack.append((registers[r], r))
+                else:
+                    stack.append((Padding('<pad %s>' % r), r))
+
+            for slot in range(moved, gadget.move, context.bytes):
+                left = gadget.move - slot
+                stack.append((Padding('<pad %#x>' % left), 'stack padding'))
+
+        return stack
 
     def resolve(self, resolvable):
         """Resolves a symbol to an address
@@ -448,7 +635,7 @@ class ROP(object):
                 if resolvable in elf.symbols:
                     return elf.symbols[resolvable]
 
-        if isinstance(resolvable, (int, long)):
+        if isinstance(resolvable, six.integer_types):
             return resolvable
 
     def unresolve(self, value):
@@ -483,23 +670,27 @@ class ROP(object):
         >>> len(val)
         15
         >>> rop.generatePadding(0,0)
-        ''
+        b''
 
         """
+
+        # Ensure we don't generate a cyclic pattern which contains badchars
+        alphabet = b''.join(packing.p8(c) for c in bytearray(string.ascii_lowercase.encode()) if c not in self._badchars)
+
         if count:
-            return cyclic.cyclic(offset + count)[-count:]
-        return ''
+            return cyclic(offset + count, alphabet=alphabet)[-count:]
+        return b''
 
     def describe(self, object):
         """
         Return a description for an object in the ROP stack
         """
-        if isinstance(object, (int, long)):
-            return self.unresolve(object)
-        if isinstance(object, str):
-            return repr(object)
-        if isinstance(object, Call):
+        if isinstance(object, (Call, constants.Constant)):
             return str(object)
+        if isinstance(object, six.integer_types):
+            return self.unresolve(object)
+        if isinstance(object, (bytes, six.text_type)):
+            return repr(object)
         if isinstance(object, Gadget):
             return '; '.join(object.insns)
 
@@ -541,15 +732,17 @@ class ROP(object):
 
             # Integers can just be added.
             # Do our best to find out what the address is.
-            if isinstance(slot, (int, long)):
+            if isinstance(slot, six.integer_types):
                 stack.describe(self.describe(slot))
                 stack.append(slot)
 
 
             # Byte blobs can also be added, however they must be
             # broken down into pointer-width blobs.
-            elif isinstance(slot, (str, unicode)):
+            elif isinstance(slot, (bytes, six.text_type)):
                 stack.describe(self.describe(slot))
+                if not isinstance(slot, bytes):
+                    slot = slot.encode()
                 slot += self.generatePadding(stack.next, len(slot) % context.bytes)
 
                 for chunk in lists.group(context.bytes, slot):
@@ -575,13 +768,17 @@ class ROP(object):
                 stack.describe(self.describe(slot))
 
                 registers    = dict(zip(slot.abi.register_arguments, slot.args))
-                setRegisters = self.setRegisters(registers)
 
-                for register, gadgets in setRegisters.items():
-                    value       = registers[register]
-                    description = self.describe(value) or 'arg%i' % slot.args.index(value)
-                    stack.describe('set %s = %s' % (register, description))
-                    stack.extend(gadgets)
+                for value, name in self.setRegisters(registers):
+                    if name in registers:
+                        index = slot.abi.register_arguments.index(name)
+                        description = self.describe(value) or repr(value)
+                        stack.describe('[arg%d] %s = %s' % (index, name, description))
+                    elif isinstance(name, Gadget):
+                        stack.describe('; '.join(name.insns))
+                    elif isinstance(name, str):
+                        stack.describe(name)
+                    stack.append(value)
 
                 if address != stack.next:
                     stack.describe(slot.name)
@@ -599,24 +796,33 @@ class ROP(object):
                 # properly, but likely also need to adjust the stack past the
                 # arguments.
                 if slot.abi.returns:
-                    if remaining:
-                        fix_size  = (1 + len(stackArguments))
-                        fix_bytes = fix_size * context.bytes
-                        adjust   = self.search(move = fix_bytes)
 
-                        if not adjust:
-                            log.error("Could not find gadget to adjust stack by %#x bytes" % fix_bytes)
+                    # Save off the address of the next gadget
+                    if remaining or stackArguments:
+                        nextGadgetAddr = stack.next
 
-                        nextGadgetAddr = stack.next + adjust.move
+                    # If there were arguments on the stack, we need to stick something
+                    # in the slot where the return address goes.
+                    if len(stackArguments) > 0:
+                        if remaining:
+                            fix_size  = (1 + len(stackArguments))
+                            fix_bytes = fix_size * context.bytes
+                            adjust   = self.search(move = fix_bytes)
 
-                        stack.describe('<adjust: %s>' % self.describe(adjust))
-                        stack.append(adjust.address)
+                            if not adjust:
+                                log.error("Could not find gadget to adjust stack by %#x bytes" % fix_bytes)
 
-                        for pad in range(fix_bytes, adjust.move, context.bytes):
-                            stackArguments.append(Padding())
-                    else:
-                        stack.describe('<pad>')
-                        stack.append(Padding())
+                            nextGadgetAddr += adjust.move
+
+                            stack.describe('<adjust @%#x> %s' % (nextGadgetAddr, self.describe(adjust)))
+                            stack.append(adjust.address)
+
+                            for pad in range(fix_bytes, adjust.move, context.bytes):
+                                stackArguments.append(Padding())
+
+                        # We could not find a proper "adjust" gadget, but also didn't need one.
+                        else:
+                            stack.append(Padding("<return address>"))
 
 
                 for i, argument in enumerate(stackArguments):
@@ -640,24 +846,24 @@ class ROP(object):
         start = base
         end   = stack.next
         size  = (stack.next - base)
+        slot_address = base
         for i, slot in enumerate(stack):
-            slot_address = stack.address + (i * context.bytes)
-            if isinstance(slot, (int, long)):
+            if isinstance(slot, six.integer_types):
                 pass
 
-            elif isinstance(slot, (str, unicode)):
+            elif isinstance(slot, (bytes, six.text_type)):
                 pass
 
             elif isinstance(slot, AppendedArgument):
                 stack[i] = stack.next
-                stack.extend(slot.resolve(stack.next))
+                stack.extend(slot.resolve(stack.next + len(slot) - context.bytes))
 
             elif isinstance(slot, CurrentStackPointer):
                 stack[i] = slot_address
 
             elif isinstance(slot, Padding):
                 stack[i] = self.generatePadding(i * context.bytes, context.bytes)
-                stack.describe('<pad>', slot_address)
+                stack.describe(slot.name, slot_address)
 
             elif isinstance(slot, Gadget):
                 stack[i] = slot.address
@@ -669,11 +875,13 @@ class ROP(object):
             else:
                 pass
 
+            slot_address += _slot_len(slot)
+
         return stack
 
 
     def find_stack_adjustment(self, slots):
-        self.search(move=slots * context.arch)
+        self.search(move=slots * context.bytes)
 
     def chain(self):
         """Build the ROP chain
@@ -801,13 +1009,13 @@ class ROP(object):
         >>> rop.raw('AAAAAAAA')
         >>> rop.raw('BBBBBBBB')
         >>> rop.raw('CCCCCCCC')
-        >>> print rop.dump()
-        0x0000:           'AAAA' 'AAAAAAAA'
-        0x0004:           'AAAA'
-        0x0008:           'BBBB' 'BBBBBBBB'
-        0x000c:           'BBBB'
-        0x0010:           'CCCC' 'CCCCCCCC'
-        0x0014:           'CCCC'
+        >>> print(rop.dump())
+        0x0000:          b'AAAA' 'AAAAAAAA'
+        0x0004:          b'AAAA'
+        0x0008:          b'BBBB' 'BBBBBBBB'
+        0x000c:          b'BBBB'
+        0x0010:          b'CCCC' 'CCCCCCCC'
+        0x0014:          b'CCCC'
         """
         if self.migrated:
             log.error('Cannot append to a migrated chain')
@@ -831,30 +1039,46 @@ class ROP(object):
             log.error('Cannot find the gadgets to migrate')
         self.migrated = True
 
-    def __str__(self):
+    def __bytes__(self):
         """Returns: Raw bytes of the ROP chain"""
         return self.chain()
 
-    def __get_cachefile_name(self, elf):
-        basename = os.path.basename(elf.file.name)
-        sha256 = hashlib.sha256(elf.get_data()).hexdigest()
-        cachedir = os.path.join(tempfile.gettempdir(), 'pwntools-rop-cache')
+    def __str__(self):
+        return str(self.chain())
+
+    def __get_cachefile_name(self, files):
+        """Given an ELF or list of ELF objects, return a cache file for the set of files"""
+        cachedir = os.path.join(tempfile.gettempdir(), 'pwntools-rop-cache-%d.%d' % sys.version_info[:2])
         if not os.path.exists(cachedir):
             os.mkdir(cachedir)
-        return os.path.join(cachedir, sha256)
+
+        if isinstance(files, ELF):
+            files = [files]
+
+        sha256 = hashlib.sha256()
+        for elf_data in sorted(elf.get_data() for elf in files):
+            sha256.update(elf_data)
+
+        return os.path.join(cachedir, sha256.hexdigest())
+
+    @staticmethod
+    def clear_cache():
+        """Clears the ROP gadget cache"""
+        cachedir = os.path.join(tempfile.gettempdir(), 'pwntools-rop-cache')
+        shutil.rmtree(cachedir)
 
     def __cache_load(self, elf):
         filename = self.__get_cachefile_name(elf)
         if not os.path.exists(filename):
             return None
-        log.info_once('Loaded cached gadgets for %r' % elf.file.name)
-        gadgets = eval(file(filename).read())
+        gadgets = eval(open(filename).read())
         gadgets = {k - elf.load_addr + elf.address:v for k, v in gadgets.items()}
+        log.info_once('Loaded %s cached gadgets for %r', len(gadgets), elf.file.name)
         return gadgets
 
     def __cache_save(self, elf, data):
         data = {k + elf.load_addr - elf.address:v for k, v in data.items()}
-        file(self.__get_cachefile_name(elf), 'w+').write(repr(data))
+        open(self.__get_cachefile_name(elf), 'w+').write(repr(data))
 
     def __load(self):
         """Load all ROP gadgets for the selected ELF files"""
@@ -871,7 +1095,7 @@ class ROP(object):
         #
 
         pop   = re.compile(r'^pop (.{3})')
-        add   = re.compile(r'^add .sp, (\S+)$')
+        add   = re.compile(r'^add [er]sp, (\S+)$')
         ret   = re.compile(r'^ret$')
         leave = re.compile(r'^leave$')
         int80 = re.compile(r'int +0x80')
@@ -950,6 +1174,11 @@ class ROP(object):
         }[context.bytes]
 
         for addr, insns in gadgets.items():
+
+            # Filter out gadgets by address against badchars
+            if set(pack(addr)) & self._badchars:
+                continue
+
             sp_move = 0
             regs = []
             for insn in insns:
@@ -999,7 +1228,10 @@ class ROP(object):
         regs = set(regs or ())
 
         for addr, gadget in self.gadgets.items():
-            if gadget.move < move:          continue
+            addr_bytes = set(pack(gadget.address))
+            if addr_bytes & self._badchars:     continue
+            if gadget.insns[-1] != 'ret':        continue
+            if gadget.move < move:               continue
             if not (regs <= set(gadget.regs)):   continue
             yield gadget
 
